
/* Maintained and distributed by Alex Bodnaru under the terms of the
   GNU Library General Public License.

   Created by, and Copyright (C) 2002 Alex Bodnaru.

This file is part of libgrx, a portable library for 2D graphics written in c

libextra is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation (version 2).

libextra is distributed "AS IS" in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with libextra; see the file COPYING.  If not, write
to the Free Software Foundation, 675 Mass Ave, Cambridge,
MA 02139, USA.

Report problems and direct all questions to:

  alexbodn@012.net.il
*/


#include <conio.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <grx20.h>
#include <grxkeys.h>
#if 0
#include <grxtext/pcfont.h>
#include <font/fonts.h>
#include <font/ramfont.h>
#endif

static struct text_info ti; /*current text info*/
GrFont *grxio_font = NULL; /*the grx font*/
GrTextOption txo; /*text options struct*/
int charw = 0, charh = 0; /*characters dimensions*/
GrCursor *normal_cursor = NULL, *solid_cursor = NULL; /*'_', '|'*/
GrCursor *cursor = NULL;
unsigned short *text_chars = NULL; /*of size screenwidth * screenheight*/
GrKeyType to_ungetch = 0, byte2 = 0;
int has_to_ungetch = 0;
int blinking = 1;
GrColor palette[17] = {16};

void    blinkvideo()
{
   blinking = 1;
}
char *  cgets(char *_str)
{
   if (_str != NULL)
   {
      int toread = *_str;
      int read = 0;
      int ch;
      while (toread && (ch = getche()) && ch != '\n' && ch != '\r')
      {
         if (ch == '\b')
         {
            --read;
            ++toread;
         }
         else
         {
            *(_str + 1 + ++read) = ch;
            --toread;
         }
      }
      *(unsigned char*)(_str + 1) = (unsigned char)read;
      *(_str + 2 + read) = '\0';
      return _str + 2;
   }
   else return NULL;
}
void    cleartext(int _left, int _top, int _right, int _bottom)
{
   if (_left > 0 && _left <= ti.screenwidth &&
       _top > 0 && _top <= ti.screenheight &&
       _right >= _left && _right <= ti.screenwidth &&
       _bottom >= _top && _bottom <= ti.screenheight)
   {
      int w = _right - _left + 1;
      int h = _bottom - _top + 1;
      int x, y;
      GrColor bg;
      unsigned short attr = ti.attribute << 8;

      for (y = 0; y < h; ++y)
      {
          for (x = 0; x < w; ++x)
            text_chars[(y + _top - 1) * ti.screenwidth + (x + _left - 1)] = attr;
      }
      bg = palette[1 + ((ti.attribute >> 4) & (blinking ? 0x07 : 0x0f))];
      if (cursor != NULL)
         GrEraseCursor(cursor);
      GrFilledBox((_left - 1) * charw, (_top - 1) * charh,
                  (_left + w - 1) * charw - 1, (_top + h - 1) * charh - 1,
                  bg);
      if (cursor != NULL)
         GrDisplayCursor(cursor);
   }
}
void showtext(int _left, int _top, int _right, int _bottom)
{
   if (_left > 0 && _left <= ti.screenwidth &&
       _top > 0 && _top <= ti.screenheight &&
       _right >= _left && _right <= ti.screenwidth &&
       _bottom >= _top && _bottom <= ti.screenheight)
   {
      int w = _right - _left + 1;
      int h = _bottom - _top + 1;
      int x = 0, y, cho;
      if (cursor != NULL)
         GrEraseCursor(cursor);
      for (y = 0; y < h; ++y)
      {
          cho = (x + _left - 1) + (y + _top - 1) * ti.screenwidth;
          GrDrawString(text_chars + cho, w,
                       charw * (_left - 1), charh * (y + _top - 1),
                       &txo);
      }
      if (cursor != NULL)
         GrDisplayCursor(cursor);
   }
}
void    clreol(void)
{
   cleartext(ti.curx, ti.cury, ti.winright, ti.cury);
}
void    clrscr(void)
{
   cleartext(ti.winleft, ti.wintop, ti.winright, ti.winbottom);
   gotoxy(1, 1);
}
#ifdef DJGPP
int     _conio_kbhit()
#else
int     kbhit()
#endif
{
   return has_to_ungetch || GrKeyPressed() != 0;
}
int     cputs(const char *_str)
{
   if (_str != NULL)
   {
      /*move cursor at the end only*/
      int cm = ti.currmode;
      if (cm != _NOCURSOR)
         _setcursortype(_NOCURSOR);
      for (; *_str != '\0'; ++_str)
         putch(*_str);
      if (cm != _NOCURSOR)
         _setcursortype(cm);
      return 0;
   }
   else return -1;
}

#ifdef DJGPP
   /*the djgpp implementation of cscanf and cprintf are calling
     getche, ungetch, cputs, as defined here
     */
/*
 * The next two functions are needed by cscanf
 */
static int
_scan_getche(FILE *fp)
{
  return(getche());
}

static int
_scan_ungetch(int c, FILE *fp)
{
  return(ungetch(c));
}


int
cscanf(const char *fmt, ...)
{
  return(_doscan_low(NULL, _scan_getche, _scan_ungetch, 
		     fmt, (void **) (char **)(&fmt)+1));
}

int
cprintf(const char *fmt, ...)
{
  int     cnt;
  char    buf[ 2048 ];		/* this is buggy, because buffer might be too small. */
  va_list ap;
  
  va_start(ap, fmt);
  cnt = vsprintf(buf, fmt, ap);
  va_end(ap);
  
  cputs(buf);
  return cnt;
}

#else /*not implemented*/
int     cscanf(const char *_format, ...)
{
    return 0;
    _format = NULL;
}

int
cprintf(const char *_format, ...)
{
    return 0;
    _format = NULL;
}
#endif
void    delline(void)
{
   if (ti.cury >= ti.wintop && ti.cury <= ti.winright)
   {
       movetext(ti.winleft, ti.cury + 1, ti.winright, ti.winbottom,
                ti.winleft, ti.cury);
       cleartext(ti.winleft, ti.winbottom, ti.winright, ti.winbottom);
   }
}
int     getch()
{
   GrKeyType saved = 0;
   if (has_to_ungetch)
   {
      saved = to_ungetch;
      has_to_ungetch = 0;
   }
   else if (byte2 != 0)
   {
      saved = byte2;
      byte2 = 0;
   }
   else
   {
      saved = GrKeyRead();
      if (saved > 0xff)
      {
         byte2 = saved & 0xff;
         saved = 0;
      }
   }
   return saved;
}
int     getche(void)
{
   GrKeyType got = getch();
   if (got > GrKey_NoKey && got < GrKey_OutsideValidRange)
      putch(got);
   return got;
}
int     gettext(int _left, int _top, int _right, int _bottom, void *_destin)
{
   /*get all or nothing*/
   if (_left > 0 && _left <= ti.screenwidth &&
       _top > 0 && _top <= ti.screenheight &&
       _right >= _left && _right <= ti.screenwidth &&
       _bottom >= _top && _bottom <= ti.screenheight &&
       _destin != NULL)
   {
      int w = _right - _left + 1;
      int h = _bottom - _top + 1;
      int x, y;
      unsigned short *_dst = _destin;
      for (y = 0; y < h; ++y)
      {
        for (x = 0; x < w; ++x)
        {
            *(_dst + y * w + x) =
                *(text_chars + (y + _top - 1) * ti.screenwidth + x + _left - 1);
        }
      }
      return 1;
   }
   return 0;
}
void    gettextinfo(struct text_info *_r)
{
   if (_r != NULL)
      *_r = ti;
}
void    gotoxy(int _x, int _y)
{
   /*should be inside window*/
   if (_x >= 1 && _x <= ti.winright - ti.winleft + 1 &&
       _y >= 1 && _y <= ti.winbottom - ti.wintop + 1 &&
       (ti.curx != _x + ti.winleft - 1 || ti.cury != _y + ti.wintop - 1))
   {
      ti.curx = _x + ti.winleft - 1;
      ti.cury = _y + ti.wintop - 1;
      if (cursor != NULL)
         GrMoveCursor(cursor, (ti.curx - 1) * charw, (ti.cury - 1) * charh);
   }
}
void    gppconio_init(void) /*mode and font should be given as parameters*/
{
#if 1
   grxio_font = &GrDefaultFont;
   charw = grxio_font->maxwidth;
   charh = grxio_font->h.height;
#else
   charw = 8;
   charh = getramfontheight();
   grxio_font = GrLoadPcFont("pcfont", "ram", NULL, charh, charw, 256, 0);
#endif
/*   if (GrSetMode(GR_width_height_color_graphics, 1024, 768, 256))*/
   if (GrSetMode(GR_default_graphics))
   {
       char *cursorbmp = NULL; /*the cursor bitmap*/
       int c;
       GrColor *egacolors = GrAllocEgaColors();
       for (c = 0; c < 16; ++c)
           palette[c + 1] = egacolors[c];
       ti.attribute = ti.normattr = WHITE | (BLACK << 4);
       ti.winleft = ti.wintop = 1;
       ti.curx = ti.cury = 0;
       ti.winright = ti.screenwidth = GrSizeX() / charw;
       ti.winbottom = ti.screenheight = GrSizeY() / charh;
       ti.currmode = _NORMALCURSOR;
       txo.txo_font = grxio_font;
       txo.txo_fgcolor.p = palette;
       txo.txo_bgcolor.p = palette;
       txo.txo_direct = GR_TEXT_RIGHT;
       txo.txo_xalign = GR_ALIGN_LEFT;
       txo.txo_yalign = GR_ALIGN_TOP;
       txo.txo_chrtype = GR_ATTR_TEXT;
       cursorbmp = calloc(sizeof(char), charw * charh);
       if (cursorbmp != NULL)
       {
           GrColor msc[2] = {1, 0};
           msc[1] = egacolors[WHITE];
           memset(cursorbmp + charw * (charh - 2 * grxio_font->h.ulheight),
                  '\1', charw * 2 * grxio_font->h.ulheight);
           normal_cursor = GrBuildCursor(cursorbmp, charw, charw, charh, 0, 0,
                                         msc);
           memset(cursorbmp, 1, charw * charh);
           solid_cursor = GrBuildCursor(cursorbmp, charw, charw, charh, 0, 0,
                                        msc);
           free(cursorbmp);
       }
       else exit(1);
       to_ungetch = byte2 = 0;
       has_to_ungetch = 0;
       blinking = 1;
       text_chars =
           malloc(sizeof(unsigned short) * ti.screenwidth * ti.screenheight);
       cursor = normal_cursor;
       GrDisplayCursor(cursor);
       clrscr();
   }
   else exit(1);
}
void    highvideo(void)
{
   ti.attribute |= 0x40;
}
void    insline(void)
{
   if (ti.cury >= ti.wintop && ti.cury <= ti.winright)
   {
       movetext(ti.winleft, ti.cury, ti.winright, ti.winbottom - 1,
                ti.winleft, ti.cury + 1);
       cleartext(ti.winleft, ti.cury, ti.winright, ti.cury);
   }
}
void    intensevideo()
{
   blinking = 0;
}
void    lowvideo(void)
{
   ti.attribute &= ~0x40;
}
int     movetext(int _left, int _top, int _right, int _bottom, int _destleft, int _desttop)
{
   /*move all or nothing*/
   if (_left >= 1 && _left <= ti.screenwidth &&
       _top >= 1 && _top <= ti.screenheight &&
       _right >= _left && _right <= ti.screenwidth &&
       _bottom >= _top && _bottom <= ti.screenheight &&
       _destleft >= 1 && _destleft <= ti.screenwidth &&
       _desttop >= 1 && _desttop <= ti.screenheight)
   {
      int w = _right - _left + 1;
      int h = _bottom - _top + 1;
      int y, yb, ye, ydelta;
      int x, xb, xe, xdelta;

      if (_destleft + w - 1 > ti.screenwidth)
         return 0;
      if (_desttop + h - 1 > ti.screenheight)
         return 0;
      if (_desttop > _top)
         yb = h - 1, ye = -1, ydelta = -1;
      else yb = 0, ye = h, ydelta = 1;
      --yb, --ye;
      if (_destleft > _left)
         xb = w - 1, xe = -1, xdelta = -1;
      else xb = 0, xe = w, xdelta = 1;
      --xb, --xe;
      for (y = yb; y != ye; y += ydelta)
      {
          for (x = xb; x != xe; x += xdelta)
          {
              text_chars[(y + _desttop) * ti.screenwidth + _destleft + x] =
                  text_chars[(y + _top) * ti.screenwidth + _left + x];
          }
      }
#if 1 /*graphically much faster*/
      if (cursor != NULL)
         GrEraseCursor(cursor);
      GrBitBlt(NULL, (_destleft - 1) * charw, (_desttop - 1) * charh,
               NULL, (_left - 1) * charw, (_top - 1) * charh,
               (_left + w - 1) * charw - 1, (_top + h - 1) * charh - 1,
               GrWRITE);
      if (cursor != NULL)
         GrDisplayCursor(cursor);
#else
      showtext(_destleft, _desttop, _destleft + w - 1, _desttop + h - 1);
#endif
      return 1;
   }
   return 0;
}
void    normvideo(void)
{
   ti.attribute = ti.normattr;
}
int     putch(int _c)
{
   int _x = ti.curx, _y = ti.cury;
   if (_c == '\r')
      _x = ti.winleft;
   else if (_c == '\n')
   {
      if (_y == ti.winbottom)
      {
         movetext(ti.winleft, ti.wintop + 1, ti.winright, ti.winbottom, ti.winleft, ti.wintop);
         cleartext(ti.winleft, ti.winbottom, ti.winright, ti.winbottom);
      }
      else ++_y;
   }
   else if (_c == '\b')
   {
      if (_x == ti.winleft)
      {
         if (_y > ti.wintop)
         {
            --_y;
            _x = ti.winright;
         }
      }
      else if (_x > ti.winleft)
         --_x;
      cleartext(_x, _y, _x, _y);
   }
   else if (_c != '\a')
   {
      unsigned short ach = _c | (ti.attribute << 8);
      puttext(_x, _y, _x, _y, &ach);
      if (_x == ti.winright)
      {
         putch('\r');
         putch('\n');
         _x = ti.curx, _y = ti.cury;
      }
      else ++_x;
   }
   else putchar('\a');
   gotoxy(_x - ti.winleft + 1, _y - ti.wintop + 1);
   return _c;
}
int     puttext(int _left, int _top, int _right, int _bottom, void *_source)
{
   /*put all or nothing*/
   if (_left >= 1 && _left <= ti.screenwidth &&
       _top >= 1 && _top <= ti.screenheight &&
       _right >= _left && _right <= ti.screenwidth &&
       _bottom >= _top && _bottom <= ti.screenheight &&
       _source != NULL)
   {
      int w = _right - _left + 1;
      int h = _bottom - _top + 1;
      int x, y;
      unsigned short *_src = _source;
      for (y = 0; y < h; ++y)
      {
          for (x = 0; x < w; ++x)
          {
              *(text_chars + (y + _top - 1) * ti.screenwidth + x + _left - 1) =
                  *(_src + y * w + x);
          }
      }
      showtext(_left, _top, _left + w - 1, _top + h - 1);
      return 1;
   }
   return 0;
}
void    _setcursortype(int _type)
{
    GrCursor *ocursor = cursor;
    if (ti.currmode == _type)
       return;
    if (_type == _NORMALCURSOR)
       cursor = normal_cursor;
    else if (_type == _SOLIDCURSOR)
       cursor = solid_cursor;
    else if (_type == _NOCURSOR)
       cursor = NULL;
    else return;
    ti.currmode = _type;
    if (ocursor != NULL)
       GrEraseCursor(ocursor);
    if (cursor != NULL)
    {
       GrDisplayCursor(cursor);
       GrMoveCursor(cursor, (ti.curx - 1) * charw, (ti.cury - 1) * charh);
    }
}
void    _set_screen_lines(int _nlines) /*implement later*/
{
   _nlines = 0;
}
void    textattr(int _attr)
{
   ti.attribute = (unsigned char)_attr;
}
void    textbackground(int _color)
{
   ti.attribute &= 0x0f;
   _color &= (blinking ? 0x07 : 0x0f);
   ti.attribute |= (_color << 4);
}
void    textcolor(int _color)
{
   ti.attribute &= 0xf0;
   ti.attribute |= (_color & (blinking ? 0x8f : 0x0f));
}
void    textmode(int _mode) /*implement later*/
{
   _mode = 0;
}
int     ungetch(int _ch)
{
   if (has_to_ungetch)
      return EOF;
   to_ungetch = _ch;
   has_to_ungetch = 1;
   return to_ungetch;
}
int     wherex()
{
   return ti.curx - ti.winleft + 1;
}
int     wherey()
{
   return ti.cury - ti.wintop + 1;
}
void    window(int _left, int _top, int _right, int _bottom)
{
   if (_left >= 1 && _left <= ti.screenwidth &&
       _top >= 1 && _top <= ti.screenheight &&
       _right >= _left && _right <= ti.screenwidth &&
       _bottom >= _top && _bottom <= ti.screenheight)
   {
      ti.winleft = _left;
      ti.wintop = _top;
      ti.winright = _right;
      ti.winbottom = _bottom;
      gotoxy(1,1);
   }
}

